﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;

namespace XNA4ModelAnimation
{
    public class CurveModelAnimationFrame
    {
        public Vector3 Position { get; private set; }
        public TimeSpan Time { get; private set; }

        public CurveModelAnimationFrame(Vector3 Position, TimeSpan Time)
        {
            this.Position = Position;
            this.Time = Time;
        }
    }

    public class CatmullRomCurveModelAnimation
    {
        List<CurveModelAnimationFrame> frames = new List<CurveModelAnimationFrame>();
        bool loop;
        TimeSpan elapsedTime = TimeSpan.FromSeconds(0);

        Vector3 PreviousPosition;
        public Vector3 Position { get; private set; }
        public Quaternion Rotation { get; private set; }
        float heading = 0.0f;

        public CatmullRomCurveModelAnimation(List<CurveModelAnimationFrame> Frames, bool Loop)
        {
            this.frames = Frames;
            this.loop = Loop;
            Position = Frames[0].Position;
            PreviousPosition = Position;
        }

        public void Update(TimeSpan Elapsed)
        {
            // Update the time
            this.elapsedTime += Elapsed;

            TimeSpan totalTime = elapsedTime;
            TimeSpan end = frames[frames.Count - 1].Time;

            if (loop) // Loop around the total time if necessary
            {
                while (totalTime > end)
                    totalTime -= end;
            }
            else
            {
                if (totalTime > end)
                {
                    Position = frames[frames.Count - 1].Position;
                    Rotation = Quaternion.CreateFromAxisAngle(Vector3.Up, heading);
                    return;
                }
            }

            int i = 0;
            // Find the index of the current frame
            while (frames[i + 1].Time < totalTime)
                i++;
            // Find the time since the beginning of this frame
            totalTime -= frames[i].Time;

            // Find how far we are between the current and next frame (0 to 1)
            float fraction = (float)((totalTime.TotalSeconds) /
               (frames[i + 1].Time - frames[i].Time).TotalSeconds);

            // Interpolate position and rotation values between frames
            Position = catmullRom3D(frames[wrap(i - 1, frames.Count - 1)].Position,
                frames[wrap(i, frames.Count - 1)].Position,
                frames[wrap(i + 1, frames.Count - 1)].Position,
                frames[wrap(i + 2, frames.Count - 1)].Position, fraction);
            System.Diagnostics.Trace.WriteLine("Position=" + Position);

            Vector3 Direction = PreviousPosition - Position;
            Direction.Normalize();
            heading = (float)Math.Atan2(Direction.X, Direction.Z);
            System.Diagnostics.Trace.WriteLine("Dir=" + Direction + " heading=" + MathHelper.ToDegrees(heading));
            Rotation = Quaternion.CreateFromAxisAngle(Vector3.Up, heading);

            PreviousPosition = Position;
        }

        Vector3 catmullRom3D(Vector3 v1, Vector3 v2, Vector3 v3, Vector3 v4, float fraction)
        {
            return new Vector3(
                MathHelper.CatmullRom(v1.X, v2.X, v3.X, v4.X, fraction),
                MathHelper.CatmullRom(v1.Y, v2.Y, v3.Y, v4.Y, fraction),
                MathHelper.CatmullRom(v1.Z, v2.Z, v3.Z, v4.Z, fraction));
        }

        // Wraps the "value" argument around [0, max]
        int wrap(int value, int max)
        {
            while (value > max) value -= max;
            while (value < 0) value += max;
            return value;
        }

        Vector3 getTangent(Vector3 v1, Vector3 v2, Vector3 v3, Vector3 v4, float fraction)
        {
            float u = fraction;
            float u2 = u * u;

            // evaluate
            return v1 * 0.5f * (-3.0f * u2 + 4.0f * u - 1.0f) + v2* 0.5f * (9.0f * u2 - 10.0f * u)
                   + v3 * 0.5f * (-9.0f * u2 + 8.0f * u + 1.0f) + v4 * 0.5f * (3.0f * u2 - 2.0f * u);
        }

        // http://msdn.microsoft.com/en-us/library/windows/desktop/bb281622(v=vs.85).aspx
        public static Quaternion Exponential(Quaternion value) 
        {
            Quaternion result = new Quaternion();
            float angle = (float)Math.Sqrt((value.X * value.X) + (value.Y * value.Y) + (value.Z * value.Z)); 
            float sin = (float)Math.Sin(angle);
            if (Math.Abs(sin) >= float.Epsilon) 
            { 
                float coeff = sin / angle; 
                result.X = coeff * value.X; 
                result.Y = coeff * value.Y; 
                result.Z = coeff * value.Z; 
            } 
            else 
            { 
                result = value; 
            } 
            result.W = (float)Math.Cos(angle);
            return result;
        }

        // http://msdn.microsoft.com/en-us/library/windows/desktop/bb281635(v=vs.85).aspx
        public static Quaternion Logarithm(Quaternion value) 
        {
            Quaternion result = new Quaternion();

            if (Math.Abs(value.W) < 1.0) 
            { 
                float angle = (float)Math.Acos(value.W); 
                float sin = (float)Math.Sin(angle);
                if (Math.Abs(sin) >= float.Epsilon) 
                { 
                    float coeff = angle / sin; 
                    result.X = value.X * coeff; 
                    result.Y = value.Y * coeff; 
                    result.Z = value.Z * coeff; 
                } 
                else
                { 
                    result = value; 
                } 
            } 
            else 
            { 
                result = value; 
            } 
            result.W = 0.0f;
            return result;
        }

        // http://msdn.microsoft.com/en-us/library/windows/desktop/bb281657(v=vs.85).aspx
        public static Quaternion[] SquadSetup(Quaternion q1, Quaternion q2, Quaternion q3, Quaternion q4) 
        { 
            Quaternion value0 = (q1 + q2).LengthSquared() < (q1 - q2).LengthSquared() ? -q1 : q1;
            Quaternion value1 = q2;
            Quaternion value2 = (q2 + q3).LengthSquared() < (q2 - q3).LengthSquared() ? -q3 : q3; 
            Quaternion value3 = (q3 + q4).LengthSquared() < (q3 - q4).LengthSquared() ? -q4 : q4; 

            Quaternion value1Exp = Exponential(value1);
            Quaternion value2Exp = Exponential(value2); 

            Quaternion[] results = new Quaternion[3]; 
            results[0] = value1 * Exponential((Logarithm(value1Exp * value2) + Logarithm(value1Exp * value0)) * -0.25f); 
            results[1] = value2 * Exponential((Logarithm(value2Exp * value3) + Logarithm(value2Exp * value1)) * -0.25f); 
            results[2] = value2; 
            return results; 
        }

        // http://msdn.microsoft.com/en-us/library/windows/desktop/bb281656(v=vs.85).aspx
        static Quaternion Squad(Quaternion q1, Quaternion q2, Quaternion q3, Quaternion q4, float fraction) 
        {
            Quaternion[] ABC = SquadSetup(q1, q2, q3, q4);
            Quaternion start = Quaternion.Slerp(q1, ABC[2], fraction);
            Quaternion end = Quaternion.Slerp(ABC[0], ABC[1], fraction);
            return Quaternion.Slerp(start, end, 2.0f * fraction * (1.0f - fraction));
        } 
    }
}
